#!/usr/bin/perl

package passer;
use strict;
$^W = 1;

use nfqueue;
use NetPacket::IP qw(IP_PROTO_TCP);
use NetPacket::TCP;
use Sys::Syslog;
use AppConfig qw/:argcount/;
use Socket;
use Socket6;
use NetAddr::IP::Lite;

my @whitelist;
my @blacklist;
my @dnsbl_low;
my @dnsbl_med;
my @dnsbl_high;

my $version = "0.4.0";
my $daemon;
my $config_file;
my $debug;
my $debugstr;
my $dryrun;
my $score;
my $score_nordns;
my $score_dnsbl_low;
my $score_dnsbl_med;
my $score_dnsbl_high;
my $score_botnet_reverse;
my $score_ipinhostname_reverse;
my $score_localhost_reverse;
my $score_ip_reverse;
my $threshold;
my $filter;
my $queue;
my $queuenum;

sub cleanup()
{
	openlog("synspam",'','daemon');
	syslog('info', "synspam stopped");
	closelog();
	$queue->unbind(AF_INET);
	$queue->close();

	exit 0;
}

sub cb()
{
	my ($dummy,$payload) = @_;
	if ($payload) {

	my $ip_obj = NetPacket::IP->decode($payload->get_data());
	my $ipsrc = $ip_obj->{src_ip};

        my $result = "";
        $debugstr = " ";
        $score = 0;

	run_tests($ipsrc);

	 my $connection =  "$ipsrc score=".$score;

	 openlog("synspam",'','daemon');
	 if ($score >= $threshold) {
		if ($dryrun == 1) {
			$result = "reject ".$connection;
			$payload->set_verdict($nfqueue::NF_ACCEPT);
		}
		else {
			$result = "reject ".$connection;
			filter($payload,$filter);
		}
	 }
	 else {
		$result = "accept ".$connection;
		$payload->set_verdict($nfqueue::NF_ACCEPT);
	 }

	 if ($debug == 1) {
		 $result .= $debugstr;
	 }

	 syslog('info', "$result");
	 closelog;


	}
}

sub filter
{
   my ($payload,$filter) = @_;
   SWITCH: for ($filter) {
	/reject/ && do { $payload->set_verdict_mark($nfqueue::NF_REPEAT,1); last;};
	/drop/ && do { $payload->set_verdict($nfqueue::NF_DROP); last;};
   }
}

sub is_blacklisted {
  my ($ip) = new NetAddr::IP::Lite (@_);
  my $black_block;

  foreach $black_block (@blacklist) {
    my $block = new NetAddr::IP::Lite($black_block);
    if ($block->contains($ip)) {
        $debugstr .= "blacklisted ";
        return 1;
    }
  }
  return 0;

}


sub is_whitelisted {
  my ($ip) = new NetAddr::IP::Lite (@_);
  my $white_block;

  foreach $white_block (@whitelist) {
    my $block = new NetAddr::IP::Lite($white_block);
    if ($block->contains($ip)) {
        $debugstr .= "whitelisted ";
        return 1;
    }
  }
  return 0;

}

sub check_dnsbl {
   
   my ($ip, $score_dnsbl, $reason, @dnsbl_list) = @_;
   my $zone = "";

  foreach $zone (@dnsbl_list) {

    $ip =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/;

    my $query = "$4.$3.$2.$1.$zone";
    my $h = gethostbyname($query);
  
    if(defined($h)) {
       $score += $score_dnsbl;
       $debugstr .= $reason;
    }
  }
}

sub check_ptr_match_ip {

my ($hostname, $ip) = @_;
my $ip_ptr;

my @ptr = gethostbyname("$hostname");

    unless (defined(@ptr)) {
	    $score += $score_ip_reverse;
	    $debugstr .= "reverse_ip ";
	    return 0;
    }

@ptr = map { inet_ntoa($_) } @ptr [ 4 .. $#ptr];
foreach $ip_ptr (@ptr) {
    if ($ip_ptr eq $ip) {
	    return 0;
    }
}
    $score += $score_ip_reverse;
    $debugstr .= "reverse_ip ";
}

sub check_words {

   my ($name) = @_;
   my @clientwords = ('.*dsl.*', 'cable', 'catv', 'ddns', 'dhcp',
    'dial(-?up)?', 'dip', 'docsis', 'dyn(amic)?(ip)?', 'modem', 'ppp(oe)?',
    'res(net|ident(ial)?)?', 'bredband', 'broadband', 'triband'
    , 'client', 'fixed', 'ip', 'pool', 'static', 'user'  # controversial ones
    );
   my $wordre = '((\b|\d)' . join('(\b|\d))|((\b|\d)', @clientwords) . '(\b|\d))';
   my $wordexp = '(' . $wordre . ')\S*\.\S+\.\S+$';

   if (($name ne "") && ($wordexp ne "") && ($name =~ /(?:$wordexp)/i) ) {
      $score += $score_botnet_reverse;
      $debugstr .= "botnet_reverse ";

      return (1);
      }
   return (0);
}

sub check_localhost {

my ($name) = @_;
   if ($name =~ /localhost(.)?$/i) {
      $debugstr .= "localhost_reverse ";
      $score += $score_localhost_reverse;
      return (1);
   }
return (0);
}

sub check_ipinhostname {

   my ($name, $ip) = @_;
   my ($a, $b, $c, $d, $e, $f, $g, $h, $i, $j, $k, $l, $m, $n);

   unless ( (defined ($name)) && ($name ne "") ) { return 0; }

   ($a, $b, $c, $d) = split(/\./, $ip); # decimal octets

   $e = ($a * 256 * 256 * 256) + ($b * 256 * 256) + ($c * 256) + $d;
   $f = ($a * 256 * 256) + ($b * 256) + $c;        # first 3 octets
   $g = ($b * 256 * 256) + ($c * 256) + $d;        # last 3 octets
   $h = ($a * 256) + $b;                           # first 2 octets
   $i = ($b * 256) + $c;                           # middle 2 octets
   $j = ($c * 256) + $d;                           # last 2 octets

   $k = sprintf("%02x", $a);                       # first octet
   $l = sprintf("%02x", $b);                       # second octet
   $m = sprintf("%02x", $c);                       # third octet
   $n = sprintf("%02x", $d);                       # fourth octet

   "(?:$a.*$b|$b.*$c|$c.*$d|$n.*$m|$m.*$l|$l.*$k|$k.*$l|$l.*$m|" .
      "$m.*$n|$e|$f|$g|$h|$i|$j)" . ".*\..+\..+\$";
      
          if ( ($name =~ /(?:$a.*$b|$b.*$c|$c.*$d).*\..+\..+$/) ||
               ($name =~ /(?:$d.*$c|$c.*$b|$b.*$a).*\..+\..+$/) ||
               ($name =~ /(?:$n.*$m|$m.*$l|$l.*$k).*\..+\..+$/) ||
               ($name =~ /(?:$k.*$l|$l.*$m|$m.*$n).*\..+\..+$/) ||
               ($name =~ /(?:$e|$f|$g|$h|$i|$j).*\..+\..+$/   ) ) {
			$score += $score_ipinhostname_reverse;
			$debugstr .= "ipinhostname_reverse ";
	  }
}

sub rdns_testsuite {

    my ($ip) = @_;
    my $iaddr = inet_aton("$ip"); 
    my $hostname = gethostbyaddr($iaddr, AF_INET);

    if (defined($hostname)) {
	   check_ptr_match_ip($hostname,$ip);
           check_words($hostname);
	   check_localhost($hostname);
	   check_ipinhostname($hostname, $ip);

	   return ($hostname);
    }

  $score += $score_nordns;
  $debugstr .= "no_rdns ";
  return "";
}

sub check_config {

 my ($file) = @_;
 my $config = AppConfig->new();

 $config->define(
	 'DEBUG' => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 0 },
	 'DRYRUN' => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 1 },
	 'DAEMON' => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 1 },
	 'QUEUENUM' => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 0 },
	 'FILTER' => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 'REJECT' },
	 'THRESHOLD' => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 6 },
	 'BLACKLIST' => { ARGCOUNT => ARGCOUNT_LIST },
	 'WHITELIST' => { ARGCOUNT => ARGCOUNT_LIST },
	 'SCORE_DNSBL_LOW' => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 1 },
	 'DNSBL_LOW' => { ARGCOUNT => ARGCOUNT_LIST },
	 'SCORE_DNSBL_MED' => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 2 },
	 'DNSBL_MED' => { ARGCOUNT => ARGCOUNT_LIST },
	 'SCORE_DNSBL_HIGH' => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 3 },
	 'DNSBL_HIGH' => { ARGCOUNT => ARGCOUNT_LIST },
	 'NORDNS' => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 2 },
	 'BOTNET_REVERSE' => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 2 },
	 'IPINHOSTNAME_REVERSE' => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 2 },
	 'IP_REVERSE' => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 3 },
	 'LOCALHOST_REVERSE' => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 3 },
 );

 $config->file($file);
 $config->args();

 $debug = $config->debug;
 $dryrun = $config->dryrun;
 $daemon = $config->daemon;
 $filter = $config->filter;
 $score_nordns = $config->nordns;
 $score_dnsbl_low = $config->score_dnsbl_low;
 $score_dnsbl_med = $config->score_dnsbl_med;
 $score_dnsbl_high = $config->score_dnsbl_high;
 $score_botnet_reverse = $config->botnet_reverse;
 $score_ipinhostname_reverse = $config->ipinhostname_reverse;
 $score_localhost_reverse = $config->localhost_reverse;
 $score_ip_reverse = $config->ip_reverse;
 $threshold = $config->threshold;
 $queuenum = $config->queuenum;
 my $tmp_black = $config->blacklist();
 my $tmp_white = $config->whitelist();
 my $tmp_low = $config->dnsbl_low();
 my $tmp_med = $config->dnsbl_med();
 my $tmp_high = $config->dnsbl_high();

 foreach (@$tmp_black) {
   push (@blacklist, $_);
 }

 foreach (@$tmp_white) {
   push (@whitelist, $_);
 }

 foreach (@$tmp_low) {
   push (@dnsbl_low, $_);
 }

 foreach (@$tmp_med) {
   push (@dnsbl_med, $_);
 }

 foreach (@$tmp_high) {
   push (@dnsbl_high, $_);
 }
}

sub run_tests {
         my ($ipsrc) = @_;

	 if (is_blacklisted($ipsrc) == 1) {
            $score = $threshold;
	    return 1;
         }
	 if (is_whitelisted($ipsrc) == 1) {
	   $score = 0;
	   return 1;
	 }

	 my $reverse = rdns_testsuite($ipsrc);
	 check_dnsbl($ipsrc, $score_dnsbl_high, "dnsbl_high ", @dnsbl_high);
	 check_dnsbl($ipsrc, $score_dnsbl_med, "dnsbl_med ", @dnsbl_med);
	 check_dnsbl($ipsrc, $score_dnsbl_low, "dnsbl_low ", @dnsbl_low);
}

sub daemonize {
  chdir '/' or die "Can't chdir to /: $!";
  open STDIN, '/dev/null' or die "Can't read /dev/null: $!";
  open STDOUT, '>>/dev/null' or die "Can't write to /dev/null: $!";
  open STDERR, '>>/dev/null' or die "Can't write to /dev/null: $!";

  defined (my $pid = fork) or die "Can't fork: $!";
  exit if $pid;
  umask 0;
}

# MAIN PROGRAM

sub main {

  $queue = new nfqueue::queue();
  $queue->set_callback(\&cb);
  $queue->fast_open($queuenum, AF_INET);

  $SIG{INT} = \&cleanup;
  $SIG{HUP} = \&cleanup;
  $SIG{TERM} = \&cleanup;

  $queue->try_run() or die("Unable to \n");
}

   $| = 1; # Make output line-buffered so it will be flushed to syslog
  
   if (-e '/etc/synspam.conf') {
	$config_file = '/etc/synspam.conf';
   }
   elsif (-e '/usr/local/etc/synspam.conf') {
	$config_file ='/usr/local/etc/synspam.conf';
   }
   else {
	die "no config file";
   }
   check_config($config_file);

   openlog("synspam",'','daemon') or die("Couldn't log output stream: $!\n");
   syslog('info', "daemon version $version succesfully started");
   closelog;

   if ($daemon == 1) {
	   daemonize();
   }
   else {
	   open STDIN, '/dev/null' or die "Can't read /dev/null: $!";
	   open STDOUT, '>>/dev/null' or die "Can't write to /dev/null: $!";
	   open STDERR, '>>/dev/null' or die "Can't write to /dev/null: $!";
  }

  # Now We're ready for the main loop 
   main();

